-- scope stuff
gearquipper = gearquipper or {};
local c = gearquipper;

c.FIELD_CURRENTSET = "FIELD_CURRENTSET";
c.FIELD_SETS = "FIELD_SETS";
c.FIELD_SLOTS = "FIELD_SLOTS";
c.FIELD_ACTIONSLOTS = "FIELD_ACTIONSLOTS";
c.FIELD_TYPE = "FIELD_TYPE";
c.FIELD_ID = "FIELD_ID";
c.FIELD_SUBTYPE = "FIELD_SUBTYPE";

c.VALUE_NONE = "VALUE_NONE";

function c:LoadSetNames()
	local setNames = {};
	for setName in pairs(GQ_DATA[c.realmName][c.charName][c.FIELD_SETS]) do
		table.insert(setNames, setName);
	end

	table.sort(setNames);
	return setNames;
end

function c:LoadCurrentSetName()
	return GQ_DATA[c.realmName][c.charName][c.FIELD_CURRENTSET];
end

function c:SaveCurrentSetName(setName)
	GQ_DATA[c.realmName][c.charName][c.FIELD_CURRENTSET] = setName;
end

function c:LoadSet(setName)
	setName = setName or c:GetText("default");
	return GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName];
end

function c:LoadCurrentSet()
	local currentSetName = c:LoadCurrentSetName();
	if currentSetName then
		return GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][currentSetName];
	end
end

function c:SaveSet(setName)
	if setName then
		GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName] = GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName] or {};
		GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName][c.FIELD_SLOTS] = GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName][c.FIELD_SLOTS] or {};

		for slotId, slotName in pairs(c:GetSlotInfo()) do
			c:SaveSlot(slotId, setName, false);
		end

		c:SaveActionBars(setName);
		c:SaveCurrentSetName(setName);
		c:ResetMenuStructure();
		c:SetPaperDollLabelText();
	end
end

function c:SaveDefaultSet()
	c:SaveSet(c:GetText("default"));
end

function c:DeleteSet(setName)
	GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName] = nil;
	c:ResetMenuStructure();
end

function c:SaveSlot(slotId, setName, notify)
	setName = setName or c:LoadCurrentSetName();
	if notify == nil then
		notify = true;
	end

	local itemId = c:GetItemId(slotId);
	GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName][c.FIELD_SLOTS][slotId] = itemId;
	if notify and GQ_OPTIONS[c.OPT_NOTIFYCHANGES] then
		c:Println(c:GetText("<%s> Saved slot %s (%s): %s", setName, slotId, c:GetDisplaySlotName(slotId), c:GetItemLink(itemId)));
	end
end

function c:LoadSlotActive()
	return GQ_DATA[c.realmName][c.charName][c.FIELD_SLOTS];
end

function c:SaveSlotActive(slotId, active)
	GQ_DATA[c.realmName][c.charName][c.FIELD_SLOTS] = GQ_DATA[c.realmName][c.charName][c.FIELD_SLOTS] or {};
	GQ_DATA[c.realmName][c.charName][c.FIELD_SLOTS][slotId] = active;
end

function c:SaveActionBars(setName)
	GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName] = GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName] or {};
	GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName][c.FIELD_ACTIONSLOTS] = GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName][c.FIELD_ACTIONSLOTS] or {};

	for slotId = 1, 120 do
		c:SaveActionSlot(slotId, setName, true, false);
	end
end

function c:SaveActionSlot(slotId, setName, saveAll, notify)
	setName = setName or c:LoadCurrentSetName();
	if notify == nil then
		notify = true;
	end

	local entry = GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName][c.FIELD_ACTIONSLOTS][slotId];
	if saveAll or not c:IsSameAction(slotId, entry) then
		if HasAction(slotId) then
			local actionType, id, subType = GetActionInfo(slotId);

			local newEntry = {};
			newEntry[c.FIELD_TYPE] = actionType;
			newEntry[c.FIELD_ID] = id;
			newEntry[c.FIELD_SUBTYPE] = subType;

			GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName][c.FIELD_ACTIONSLOTS][slotId] = newEntry;
		else
			GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName][c.FIELD_ACTIONSLOTS][slotId] = c.VALUE_NONE;
		end

		if notify and GQ_OPTIONS[c.OPT_NOTIFYCHANGES] then
			local actionItemName = c:GetText("Empty");

			if GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName][c.FIELD_ACTIONSLOTS][slotId] ~= c.VALUE_NONE then
				local aType, id = GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName][c.FIELD_ACTIONSLOTS][slotId][c.FIELD_TYPE], GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName][c.FIELD_ACTIONSLOTS][slotId][c.FIELD_ID];

				if aType == "spell" then
					local name, rank = c:GetSpellName(id), c:GetSpellRank(id);
					if rank and rank ~= "" then
						actionItemName = c:GetText("%s (%s)", name, rank);
					else
						actionItemName = name;
					end
				elseif aType == "item" then
					actionItemName = c:GetRarityColor(c:GetItemRarity(id))["hex"] .. c:GetItemName(id);
				elseif aType == "macro" then
					actionItemName = c:GetMacroName(id);
				end
			end
			c:Println(c:GetText("<%s> Saved action slot %s: %s", setName, slotId, actionItemName));
		end
	end
end

function c:LoadActionBars(setName, retry)
	for slotId = 1, 120 do
		local entry = GQ_DATA[c.realmName][c.charName][c.FIELD_SETS][setName][c.FIELD_ACTIONSLOTS][slotId];

		if not c:IsSameAction(slotId, entry) then
			if InCombatLockdown() then
				-- breaks loop if combat starts
				if not retry then
					c:Println(c:GetText("Switching interrupted by combat. It will be re-attempted once combat is over."));
				end
				C_Timer.After(1, function() c:SwitchToSet(setName, true); end);
				return;
			end

			ClearCursor();
			PickupAction(slotId);
			ClearCursor();

			if entry ~= c.VALUE_NONE then
				if entry[c.FIELD_TYPE] == "spell" then
					PickupSpell(entry[c.FIELD_ID]);
				elseif entry[c.FIELD_TYPE] == "item" then
					PickupItem(entry[c.FIELD_ID]);
				elseif entry[c.FIELD_TYPE] == "macro" then
					PickupMacro(entry[c.FIELD_ID]);
				end

				PlaceAction(slotId);
				ClearCursor();
			end

			if GQ_OPTIONS[c.OPT_HIGHLIGHTCHANGES] then
				c:HighlightActionSlot(slotId);
			end
		end
	end

	return true;
end

function c:IsSameAction(slotId, entry)
	if HasAction(slotId) and entry and entry ~= c.VALUE_NONE then
		local actionType, id, subType = GetActionInfo(slotId);
		if actionType == entry[c.FIELD_TYPE] and id == entry[c.FIELD_ID] and subType == entry[c.FIELD_SUBTYPE] then
			return true;
		end
	elseif not HasAction(slotId) and (not entry or entry == c.VALUE_NONE) then
		return true;
	end
end