-- scope stuff
gearquipper = gearquipper or {};
local c = gearquipper;

local DIALOG_ADDSET = "DIALOG_ADDSET";
local DIALOG_DELETESET = "DIALOG_DELETESET";
local DIALOG_RESET = "DIALOG_RESET";

local paperDollButton, paperDollLabel, paperDollMenu;
local slotCheckboxes, quickBars;
local gqToolTip;

local secureActionButtons = {};

function c:InitUI()
	paperDollButton = paperDollButton or c:CreatePaperDollButton();
	paperDollLabel = paperDollLabel or c:CreatePaperDollLabel();
	paperDollMenu = paperDollMenu or c:CreatePaperDollMenu();

	_G["PaperDollFrame"]:HookScript("OnShow", function() c:SetSlotInfo(); end);
	_G["PaperDollFrame"]:HookScript("OnHide", function()
		c:ResetMenuStructure();
		c:CloseQuickBars();
	 end);

	c:CreateSecureActionButtons();
	--c:CreateSlotCheckboxes();
	c:CreateDialogs();
end

function c:CreateSecureActionButtons()
	for _, setName in ipairs(c:LoadSetNames()) do
		local set = c:LoadSet(setName);
		local macrotext;

		if set[INVSLOT_MAINHAND] == c.VALUE_NONE then
			macrotext = string.format("/unequipslot %s", INVSLOT_MAINHAND);
		else
			macrotext = string.format("/equipslot %s %s", INVSLOT_MAINHAND, c:GetItemName(set[c.FIELD_SLOTS][INVSLOT_MAINHAND]));
		end

		if set[INVSLOT_OFFHAND] == c.VALUE_NONE then
			macrotext = string.format("/unequipslot %s", INVSLOT_OFFHAND);
		else
			macrotext = string.format("/equipslot %s %s", INVSLOT_OFFHAND, c:GetItemName(set[c.FIELD_SLOTS][INVSLOT_OFFHAND]));
		end

		if set[INVSLOT_RANGED] == c.VALUE_NONE then
			macrotext = string.format("/unequipslot %s", INVSLOT_RANGED);
		else
			macrotext = string.format("/equipslot %s %s", INVSLOT_RANGED, c:GetItemName(set[c.FIELD_SLOTS][INVSLOT_RANGED]));
		end

		c:SetSecureAction(setName, macrotext);
	end
end

function c:CreateSecureActionButton(setName)
	if setName then
		if not secureActionButtons[setName] then
			local button = CreateFrame("Button", "GQ_SecureActionButton_" .. setName, UIParent, "SecureActionButtonTemplate");
			button:RegisterForClicks("AnyUp");
			button:SetAttribute("type", "macro");
			secureActionButtons[setName] = button;
		end
		return secureActionButtons[setName];
	end
end

function c:SetSecureAction(setName, macrotext)
	if setName and macrotext then
		local button = c:CreateSecureActionButton(setName);
		button:SetAttribute("macrotext", macrotext);
	end
end

function c:CallSecureAction(setName)
	if secureActionButtons[setName] then
		-- experimental
		--C_Timer.After(0.1, function()
		--	secureActionButtons[setName]:Click();
		--end);
	end
end

function c:CreatePaperDollButton()
	local button = CreateFrame("Button", "GQ_PaperDollButton", _G["PaperDollItemsFrame"]);
	button:SetPoint("TOPRIGHT", _G["PaperDollItemsFrame"], "TOPRIGHT", -40, -45);
	button:SetWidth(50);
	button:SetHeight(20);
	button:SetText("GQ");
	button:SetNormalFontObject("GameFontNormal");
	button:RegisterForClicks("LeftButtonUp");

	local ntex = button:CreateTexture();
	ntex:SetTexture("Interface/Buttons/UI-Panel-Button-Up");
	ntex:SetTexCoord(0, 0.625, 0, 0.6875);
	ntex:SetAllPoints();
	button:SetNormalTexture(ntex);

	local htex = button:CreateTexture();
	htex:SetTexture("Interface/Buttons/UI-Panel-Button-Highlight");
	htex:SetTexCoord(0, 0.625, 0, 0.6875);
	htex:SetAllPoints();
	button:SetHighlightTexture(htex);

	local ptex = button:CreateTexture();
	ptex:SetTexture("Interface/Buttons/UI-Panel-Button-Down");
	ptex:SetTexCoord(0, 0.625, 0, 0.6875);
	ptex:SetAllPoints();
	button:SetPushedTexture(ptex);

	local dtex = button:CreateTexture();
	dtex:SetTexture("Interface/Buttons/UI-Panel-Button-Disabled");
	dtex:SetTexCoord(0, 0.625, 0, 0.6875);
	dtex:SetAllPoints();
	button:SetDisabledTexture(dtex);

	button:SetScript("OnClick", function() c:TogglePaperDollMenu(); end);

	return button;
end

function c:DisablePaperDollButton()
	paperDollButton:Disable();
	c:EnablePaperDollButton();
end

function c:EnablePaperDollButton()
	if InCombatLockdown() then
		C_Timer.After(1, function() c:EnablePaperDollButton(); end); -- try again
	else
		paperDollButton:Enable();
	end
end

function c:CreatePaperDollLabel()
	local label = CreateFrame("Frame", "GQ_PaperDollLabel", _G["PaperDollItemsFrame"]);
	label:SetPoint("CENTER", _G["PaperDollItemsFrame"], "CENTER", 0, -5);
	label:SetWidth(200);
	label:SetHeight(20);
	label:EnableMouse();
	label:SetScript("OnEnter", function(self) self:SetAlpha(0); end);
	label:SetScript("OnLeave", function(self) self:SetAlpha(1); end);

	label.text = label:CreateFontString(nil, "ARTWORK");
	label.text:SetFont("Fonts\\ARIALN.ttf", 13, "OUTLINE");
	label.text:SetPoint("CENTER", 0, 0);

	return label;
end

function c:SetPaperDollLabelText(setName)
	setName = setName or c:LoadCurrentSetName();
	if c:IsSetComplete() then
		paperDollLabel.text:SetText(c:GetText("%sCurrent set: \"%s\"", "|cffffcc00", setName));
	else
		paperDollLabel.text:SetText(c:GetText("%sCurrent set: \"%s\"", "|cffff0000", setName));
	end
end

local menuStructure;
function c:GetMenuStructure()
	if not menuStructure then
		local switchSet, addSet, removeSet, options = c:GetText("Switch to set:"), c:GetText("Add set..."), c:GetText("Remove set"), c:GetText("Optionen");
		local setNames = c:LoadSetNames();

		menuStructure = {
			[0] = {
				switchSet,
				"separator1",
				addSet,
				removeSet,
				"separator2",
				options
			},
			[switchSet] = {},
			["separator1"] = {},
			[addSet] = {},
			[removeSet] = {},
			["separator2"] = {},
			[options] = {}
		};

		for _, setName in ipairs(setNames) do
			table.insert(menuStructure[switchSet], { ["name"] = setName;} );
			table.insert(menuStructure[removeSet], { ["name"] = setName;} );
		end

		table.insert(menuStructure[options], { ["name"] = c:GetText("Always save changes"); } );
		table.insert(menuStructure[options], { ["name"] = c:GetText("Switch action bar configuration"); } );
		table.insert(menuStructure[options], { ["name"] = c:GetText("Confirm changes in chat"); } );
		table.insert(menuStructure[options], { ["name"] = c:GetText("Highlight switched slots"); } );
		table.insert(menuStructure[options], { ["name"] = c:GetText("Reset GearQuipper"); } );
	end
	return menuStructure;
end

function c:CreatePaperDollMenu()
	local dropdown = CreateFrame("Frame", "GQ_PaperDollMenu", UIParent, "UIDropDownMenuTemplate");
	UIDropDownMenu_Initialize(dropdown, function(self, level)
		level = level or 1;

		if (level == 1) then
			for no, key in pairs(c:GetMenuStructure()[0]) do
				local subarray = c:GetMenuStructure()[key];

				local info = UIDropDownMenu_CreateInfo();
				info.hasArrow = table.getn(subarray) > 0;
				info.notCheckable = true;
				info.notClickable = c:StartsWith(key, "separator");
				if not c:StartsWith(key, "separator") then
					info.text = key;
					info.value = {
						["Level1_Key"] = key;
					};

					if key == c:GetText("Add set...") then
						info.func = function()
							StaticPopup_Show(DIALOG_ADDSET);
							c:TogglePaperDollMenu();
						end;
					end
				end
				UIDropDownMenu_AddButton(info, level);
			end
		end

		if (level == 2) then
			local Level1_Key = UIDROPDOWNMENU_MENU_VALUE["Level1_Key"];
			local subarray = c:GetMenuStructure()[Level1_Key];
			local currentSetName = c:LoadCurrentSetName();

			for key, subsubarray in pairs(subarray) do
				local info = UIDropDownMenu_CreateInfo();
				info.hasArrow = false;
				info.text = subsubarray["name"];
				info.value = {
					["Level1_Key"] = Level1_Key;
					["Sublevel_Key"] = key;
				};

				if Level1_Key == c:GetText("Switch to set:") then
					info.notCheckable = false;
					info.checked = currentSetName and currentSetName == subsubarray["name"];
					info.func = function()
						c:SwitchToSet(subsubarray["name"]);
						c:ResetMenuStructure();
					end;
				elseif Level1_Key == c:GetText("Remove set") then
					info.notCheckable = true;
					info.notClickable = false;
					info.disabled = currentSetName == subsubarray["name"];
					info.func = function()
						local dialog = StaticPopup_Show(DIALOG_DELETESET, subsubarray["name"]);
						dialog.data = subsubarray["name"];
						c:ResetMenuStructure();
					end;
				elseif Level1_Key == c:GetText("Optionen") then
					if subsubarray["name"] == c:GetText("Always save changes") then
						info.checked = GQ_OPTIONS[c.OPT_ALWAYSSAVECHANGES];
						info.func = function(self, arg1, arg2, checked)
							GQ_OPTIONS[c.OPT_ALWAYSSAVECHANGES] = not checked;
							c:ResetMenuStructure();
						end;
					elseif subsubarray["name"] == c:GetText("Switch action bar configuration") then
						info.checked = GQ_OPTIONS[c.OPT_ACTIONBARS];
						info.func = function(self, arg1, arg2, checked)
							GQ_OPTIONS[c.OPT_ACTIONBARS] = not checked;
							c:ResetMenuStructure();
						end;
					elseif subsubarray["name"] == c:GetText("Confirm changes in chat") then
						info.checked = GQ_OPTIONS[c.OPT_NOTIFYCHANGES];
						info.func = function(self, arg1, arg2, checked)
							GQ_OPTIONS[c.OPT_NOTIFYCHANGES] = not checked;
							c:ResetMenuStructure();
						end;
					elseif subsubarray["name"] == c:GetText("Highlight switched slots") then
						info.checked = GQ_OPTIONS[c.OPT_HIGHLIGHTCHANGES];
						info.func = function(self, arg1, arg2, checked)
							GQ_OPTIONS[c.OPT_HIGHLIGHTCHANGES] = not checked;
							c:ResetMenuStructure();
						end;
					elseif subsubarray["name"] == c:GetText("Reset GearQuipper") then
						info.notCheckable = true;
						info.func = function()
							c:ShowResetDialog();
							c:ResetMenuStructure();
						end;
					end
				end

				UIDropDownMenu_AddButton(info, level);
			end
		end
	end, "MENU");

	return dropdown;
end

function c:TogglePaperDollMenu()
	paperDollMenu = paperDollMenu or c:CreatePaperDollMenu();
	ToggleDropDownMenu(1, nil, paperDollMenu, paperDollButton, -20, 0);
end

function c:ResetMenuStructure()
	if paperDollMenu then
		paperDollMenu:Hide();
	end
	menuStructure = nil;
	paperDollMenu = nil;
end

function c:CreateSlotCheckboxes()
	if not slotCheckboxes then
		slotCheckboxes = {};
		for slotId, slotName in pairs(c:GetSlotInfo()) do
			slotCheckboxes[slotId] = c:CreateSlotCheckbox(slotName, _G["Character" .. slotName]);
		end
	end
end

function c:CreateSlotCheckbox(slotName, parent)
	local checkbutton = CreateFrame("CheckButton", slotName .. "Checkbox", parent, "UICheckButtonTemplate");
	checkbutton:SetWidth(24);
	checkbutton:SetHeight(24);
	checkbutton:SetPoint("BOTTOMLEFT", -4, -4);
	checkbutton:RegisterForClicks("LeftButtonUp");
	checkbutton:SetScript("OnClick",
		function(self)
			local slotId = c:GetSlotId(slotName);
			c:SaveSlotActive(slotId, self:GetChecked());

			local currentSet = c:LoadCurrentSet();
			if c:GetItemId(slotId) ~= currentSet[c.FIELD_SLOTS][slotId] then
				c:SaveSlot(slotId);
				c:SetSlotInfo(slotId, "green");
			end
		end);
	return checkbutton;
end

function c:SetCheckboxValues(overrideValue, save)
	if overrideValue ~= nil then
		for slotId, checkBox in pairs(slotCheckboxes) do
			c:SetCheckboxValue(slotId, overrideValue, save);
		end
	else
		local activeSlots = c:LoadSlotActive();
		if activeSlots then
			for slotId, value in pairs(activeSlots) do
				c:SetCheckboxValue(slotId, value);
			end
		end
	end
end

function c:SetCheckboxValue(slotId, value, save)
	slotCheckboxes[slotId]:SetChecked(value);
	if save then
		c:SaveSlotActive(slotId, value)
	end
end

function c:IsSlotActive(slotId)
	return slotCheckboxes[slotId]:GetChecked();
end

function c:CreateDialogs()
	StaticPopupDialogs[DIALOG_ADDSET] = StaticPopupDialogs[DIALOG_ADDSET] or {
		text = c:GetText("Enter a name for this set:"),
		button1 = c:GetText("Apply"),
		button2 = c:GetText("Cancel"),
		hasEditBox = true,
		maxLetters = 30,
		EditBoxOnEnterPressed = function (self)
			local parent = self:GetParent();
			local setName = c:Trim(self:GetText());
			if setName and not tContains(c:LoadSetNames(), setName) then
				c:SaveSet(setName, true);
			end
			parent:Hide();
		end,
		EditBoxOnEscapePressed = function (self)
			self:GetParent():Hide();
		end,
		OnAccept = function(self, data, data2)
			local setName = c:Trim(self.editBox:GetText());
			if setName and not tContains(c:LoadSetNames(), setName) then
				c:SaveSet(setName, true);
			end
		end,
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
		preferredIndex = 3
		};

	StaticPopupDialogs[DIALOG_DELETESET] = StaticPopupDialogs[DIALOG_DELETESET] or {
		text = c:GetText("Do you really want to delete \"%s\"?", "%s"),
		button1 = c:GetText("Apply"),
		button2 = c:GetText("Cancel"),
		EditBoxOnEscapePressed = function (self)
			self:GetParent():Hide();
		end,
		OnAccept = function(self, data)
			if data and tContains(c:LoadSetNames(), data) then
				c:DeleteSet(data);
			end
		end,
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
		preferredIndex = 3
		};

	StaticPopupDialogs[DIALOG_RESET] = StaticPopupDialogs[DIALOG_RESET] or {
		text = c:GetText("This will reset GearQuipper to defaults. All your sets will be lost.\n\nAre you sure?"),
		button1 = c:GetText("Reset"),
		button2 = c:GetText("Cancel"),
		OnEscapePressed = function ()
			this:Hide();
		end,
		OnAccept = function()
			GQ_OPTIONS = {};
			GQ_DATA = {};
			c:ResetMenuStructure();
			c:Init();
			c:SaveDefaultSet();
			c:SetCheckboxValues(true, true);
		end,
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
		preferredIndex = 3
		};
end

function c:ShowResetDialog()
	StaticPopup_Show(DIALOG_RESET);
end

local slotSize, slotMargin = 40, 8;
function c:CreateQuickBar(slotId)
	local parent = _G["Character" .. c:GetSlotInfo()[slotId]];
	local frame = CreateFrame("Frame", "GQ_QuickBar_" .. slotId, parent);

    frame:SetHeight(slotSize + slotMargin);
	frame:SetPoint("TOPLEFT", parent, "TOPLEFT", slotSize, (slotMargin / 2));
	frame:SetFrameStrata("HIGH");

    frame.border = frame:CreateTexture("GQ_QuickBarBorderTex", "BORDER");
    frame.border:SetPoint("TOPLEFT");
	frame.border:SetTexture("Interface\\PaperDoll\\UI-Backpack-EmptySlot.blp");
	frame.border:SetHeight(slotSize + slotMargin);

	return frame;
end

function c:CreateQuickSlot(slotId, index)
	local frame = CreateFrame("Button", "GQ_QuickSlot_" .. slotId .. "_" .. index, quickBars[slotId], "GqItemTemplate");
    frame:SetWidth(slotSize);
    frame:SetHeight(slotSize);
	frame:SetPoint("TOPLEFT", quickBars[slotId], "TOPLEFT", (slotSize * (index - 1)) + (slotMargin * (index - 1)) + slotMargin, -(slotMargin / 2));
	frame:SetFrameStrata("HIGH");

	frame:SetScript("OnEnter", function(self)
		GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT");
		GameTooltip:SetHyperlink(self.link);
		GameTooltip:Show();
	end);
	frame:SetScript("OnLeave", function()
		GameTooltip:Hide();
	end);
	frame:SetScript("OnClick", function(self)
		ClearCursor();
		EquipItemByName(self:GetAttribute("item2"), slotId);
		c:CloseQuickBar(slotId);
		GameTooltip:Hide();
	end);

	return frame;
end

function c:OpenQuickBar(slotId)
	if slotId then
		local matchingItems, noSlots = c:GetMatchingItems(slotId);

		if noSlots > 0 then
			quickBars = quickBars or {};
			quickBars[slotId] = quickBars[slotId] or c:CreateQuickBar(slotId);
			quickBars[slotId]:SetWidth((slotSize * noSlots) + (slotMargin * (noSlots - 1)));
			quickBars[slotId].border:SetWidth((slotSize * noSlots) + (slotMargin * (noSlots - 1)) + (slotMargin * 2));
			quickBars[slotId]:Show();

			quickBars[slotId].slots = quickBars[slotId].slots or {};
			for index, itemEntry in ipairs(matchingItems) do
				c:AddQuickSlot(slotId, index, itemEntry);
			end

			GameTooltip:Hide();
		end
	end
end

function c:AddQuickSlot(slotId, slotIndex, itemEntry)
	local quickBar = quickBars[slotId];	
	quickBar.slots[slotIndex] = quickBar.slots[slotIndex] or c:CreateQuickSlot(slotId, slotIndex);
	local quickSlot = quickBar.slots[slotIndex];
	local texture, count, locked, quality, readable, lootable, link = GetContainerItemInfo(itemEntry[c.ITEM_BAG_ID], itemEntry[c.ITEM_SLOT_ID]);

	quickSlot.link = link;
	quickSlot:SetAttribute("item2", c:GetItemName(itemEntry[c.ITEM_ID]));
	quickSlot.icon:SetTexture(texture);
	if count > 1 then
		quickSlot.count:SetText(count);
		quickSlot.count:Show();
	else
		quickSlot.count:Hide();
	end

	if quality > 1 then
		quickSlot.glow:SetVertexColor(GetItemQualityColor(quality));
		quickSlot.glow:Show();
	else
		quickSlot.glow:Hide();
	end

	quickSlot:Show();
end

function c:CloseQuickBars()
	if quickBars then
		for slotId, quickBar in pairs(quickBars) do
			quickBar:Hide();
		end
	end
end

function c:CloseQuickBar(slotId)
	if quickBars and quickBars[slotId] then
		quickBars[slotId]:Hide();
		for index, slot in pairs(quickBars[slotId].slots) do
			slot:Hide();
		end
	end
end

local function HighlightButton(button)
	if button then
		ActionButton_ShowOverlayGlow(button);
		C_Timer.After(1, function () ActionButton_HideOverlayGlow(button); end);
	end
end

function c:HighlightInventorySlot(slotId)
	if c:GetSlotInfo()[slotId] then
		HighlightButton(_G["Character" .. c:GetSlotInfo()[slotId]]);
	end
end

function c:HighlightActionSlot(slotId)
	HighlightButton(c:GetActionButton(slotId));
end

function c:SetSlotInfo(slotId, color)
	if not slotId and not color then
		local set = c:LoadCurrentSet();
		if set then
			for slotId, slotName in pairs(c:GetSlotInfo()) do
				if c:IsSetItem(slotId, set[c.FIELD_SLOTS][slotId]) then
					c:SetSlotInfo(slotId, "green");
				else
					c:SetSlotInfo(slotId, "red");
				end
			end
		end
		return;
	end

	local slot = _G["Character" .. c:GetSlotInfo()[slotId]];
	if not slot.gqTex then
		slot.slotId = slotId;
		slot.gqTex = slot:CreateTexture(nil, "OVERLAY");
		slot.gqTex:SetAllPoints();
		slot:HookScript("OnEnter", function(self)
			c:InjectTooltip(self);
		end);

		slot:HookScript("OnLeave", function()
			if gqToolTip then
				gqToolTip:Hide();
			end
		end);
	end

	if color == "red" then
		slot.gqTex:SetColorTexture(1, 0, 0, 0.2);
		slot.gqInfo = c:GetText("|cffff0000Expected Item:|r %s", c:GetItemLink(c:LoadCurrentSet()[c.FIELD_SLOTS][slotId]));
	else
		slot.gqTex:SetColorTexture(0, 0, 0, 0);
		slot.gqInfo = nil;
	end
end

function c:InjectTooltip(slot)
	-- experimental
	if not gqToolTip then
		gqToolTip = CreateFrame("GameTooltip", "GQ_SlotTooltip", UIParent, "GameTooltipTemplate");
	end

	gqToolTip:SetOwner(slot, "ANCHOR_RIGHT");
	if slot.gqInfo then
		gqToolTip:AddLine(slot.gqInfo);
		gqToolTip:Show();

		GameTooltip:SetOwner(gqToolTip, "ANCHOR_BOTTOM");
		GameTooltip:SetHyperlink(c:GetItemLink(c:GetItemId(slot.slotId)));
	end
end